!-------------------------------------------------------------------------------

!                      Code_Saturne version 3.0.0
!                      --------------------------
! This file is part of Code_Saturne, a general-purpose CFD tool.
!
! Copyright (C) 1998-2013 EDF S.A.
!
! This program is free software; you can redistribute it and/or modify it under
! the terms of the GNU General Public License as published by the Free Software
! Foundation; either version 2 of the License, or (at your option) any later
! version.
!
! This program is distributed in the hope that it will be useful, but WITHOUT
! ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
! FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
! details.
!
! You should have received a copy of the GNU General Public License along with
! this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
! Street, Fifth Floor, Boston, MA 02110-1301, USA.

!-------------------------------------------------------------------------------

!===============================================================================
! Purpose:
! -------

!> \file cs_user_extra_operations.f90
!>
!> \brief This function is called at the end of each time step, and has a very
!>  general purpose
!>  (i.e. anything that does not have another dedicated user subroutine)
!>
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
! Arguments
!______________________________________________________________________________.
!  mode           name          role                                           !
!______________________________________________________________________________!
!> \param[in]     nvar          total number of variables
!> \param[in]     nscal         total number of scalars
!> \param[in]     nbpmax        max. number of particles allowed
!> \param[in]     nvp           number of particle-defined variables
!> \param[in]     nvep          number of real particle properties
!> \param[in]     nivep         number of integer particle properties
!> \param[in]     ntersl        number of return coupling source terms
!> \param[in]     nvlsta        number of Lagrangian statistical variables
!> \param[in]     nvisbr        number of boundary statistics
!> \param[in]     itepa         integer particle attributes
!>                                (containing cell, ...)
!> \param[in]     dt            time step (per cell)
!> \param[in]     rtp, rtpa     calculated variables at cell centers
!>                               (at current and previous time steps)
!> \param[in]     propce        physical properties at cell centers
!> \param[in]     propfa        physical properties at interior face centers
!> \param[in]     propfb        physical properties at boundary face centers
!> \param[in]     ettp, ettpa   particle-defined variables
!> \param[in]                    (at current and previous time steps)
!> \param[in]     tepa          real particle properties
!> \param[in]                    (statistical weight, ...
!> \param[in]     statis        statistic means
!> \param[in]     stativ        accumulator for variance of volume statisitics
!> \param[in]     tslagr        Lagrangian return coupling term
!> \param[in]                    on carrier phase
!> \param[in]     parbor        particle interaction properties
!> \param[in]                    on boundary faces
!_______________________________________________________________________________


subroutine cs_user_extra_operations &
 ( nvar   , nscal  ,                                              &
   nbpmax , nvp    , nvep   , nivep  , ntersl , nvlsta , nvisbr , &
   itepa  ,                                                       &
   dt     , rtpa   , rtp    , propce , propfa , propfb ,          &
   ettp   , ettpa  , tepa   , statis , stativ , tslagr , parbor )

!===============================================================================

!===============================================================================
! Module files
!===============================================================================

use paramx
use dimens, only: ndimfb
use pointe
use numvar
use optcal
use cstphy
use cstnum
use entsor
use lagpar
use lagran
use parall
use period
use ppppar
use ppthch
use ppincl
use field
use mesh
use user_module
!===============================================================================

implicit none

! Arguments

integer          nvar   , nscal
integer          nbpmax , nvp    , nvep  , nivep
integer          ntersl , nvlsta , nvisbr

integer          itepa(nbpmax,nivep)

double precision dt(ncelet), rtp(ncelet,*), rtpa(ncelet,*)
double precision propce(ncelet,*)
double precision propfa(nfac,*), propfb(ndimfb,*)
double precision ettp(nbpmax,nvp) , ettpa(nbpmax,nvp)
double precision tepa(nbpmax,nvep)
double precision statis(ncelet,nvlsta), stativ(ncelet,nvlsta-1)
double precision tslagr(ncelet,ntersl)
double precision parbor(nfabor,nvisbr)


! Local variables
double precision , allocatable, dimension(:) :: ytmp

! INSERT_VARIABLE_DEFINITIONS_HERE

!===============================================================================
integer impout,ii,icnt,npoint,jj,iel,iel1,irang1,irangv,iun

double precision ymax,xabs,yy,xyz(3),yy1,rrr,rrmin,xeps
double precision datave(15),xdudy
double precision, allocatable, dimension(:) :: averages,warr,dudy

double precision, dimension(:,:), pointer :: coefav
double precision, dimension(:,:,:), pointer :: coefbv
double precision, allocatable, dimension(:,:,:) :: grad


integer iccocg,idtcm,imom,inc,ipcmom,ipcliu,ntchr
!===============================================================================
! Initialization
!===============================================================================
allocate(grad(ncelet,3,3))
allocate(warr(ncelet))
! INSERT_ADDITIONAL_INITIALIZATION_CODE_HERE

!--writting frequency
ntchr=16

!===============================================================================
! User operations
!===============================================================================

!--  Only the first time step --
!-------------------------------
if (ntcabs.eq.ntpabs+1) then 

   ! Find the points on y
   ! Save them to avoid doing it again
   !----------------------------------

   npoint = 400
   allocate(ytmp(npoint))
   ytmp = 0.d0

   ymax   = 2.d0 !--channel height
   nymax  = 0
   iel1   = -999
   irang1 = -999

   do ii = 1, npoint
      ! x-coordinate of the node
      xyz(1) = 0.d0
      ! y-coordinate of the node
      xyz(2) = float(ii-1)/float(npoint-1)*ymax
      ! z-coordinate of the node
      xyz(3) = 0.d0                
      call findpt(ncelet,ncel,xyzcen,xyz(1),xyz(2),xyz(3),iel,irangv)
      
      if ((iel.ne.iel1).or.(irangv.ne.irang1)) then ! If the last point found is the same as the new one
         iel1   = iel
         irang1 = irangv
         
         if (irangp.eq.irangv) then
            xabs = xyzcen(2,iel)
         else
            xabs = 0.d0
         endif
         if (irangp.ge.0) then
            iun = 1        
            call parbcr(irangv,iun,xabs) ! broadcast from rank irangv to all others
         endif
         if (irangp.le.0)then
            nymax = nymax +1
            ytmp(nymax) = xabs
            write(*,*),'ytmp0=',ytmp(nymax)
         endif
      endif
   enddo

   if (irangp.ge.0) then
      ! -- broadcast nymax from rank 0 to all processors
      call parbci(0,1 ,nymax)
   endif
   write(nfecra,*) 'CELLS ON Y:',nymax
   write(*,*),nymax
   ! --Allocate memories for user variables--
   !-----------------------------------------
   call init_user_module(ncel,ncelet)
   write(nfecra,*) 'Initialization'
   if (irangp.le.0) then
      do ii=1,nymax
         ydat(ii)=ytmp(ii)
      enddo
   endif

   if (irangp.ge.0) then
      ! -- broadcast ydat from rank 0 to all others
      call parbcr(0,nymax,ydat) 
   endif

   write(nfecra,*) 'Conectivity'
   write(*,*) 'ydat=',ydat

  ! Find conectivity between cells and y data.  

   invarc = 0
     
  ! For each cell, find which is the nearest ydat.
  do iel = 1, ncel
     rrr = 1.d13
     rrmin = 1.d13
     do ii = 1, nymax
        xeps = 1.d-8
        yy = ydat(ii)
        yy1 = xyzcen(2,iel)
        rrr = sqrt((yy-yy1)**2)
        if (rrr.lt.rrmin)then
           rrmin = rrr
           corrs(iel) = ii !--which y-coordinate corresponds to iel node
        endif
     enddo
     jj = corrs(iel)
     invarc(jj) = invarc(jj) + 1
  enddo
  ! Add the total number of cells
  write(nfecra,*) 'Add the total number of cells'
  write(*,*) 'invarc=',invarc
  do ii = 1,nymax
     jj = invarc(ii)
     ! -- global sum over jj
     write(nfecra,*) 'jj1=',jj
     if(irangp.ge.0) call parcpt(jj) 
     write(nfecra,*) 'jj2=',jj
     ! -- The total number of points that are related to each y
     invarc(ii) = jj  
     write(nfecra,*) 'invarc=',invarc(ii)
  enddo  
  write(nfecra,*) 'deallocation'
  deallocate(ytmp)
endif

allocate(dudy(nymax))

! ======================================================
! Plane averaging :
! ======================================================
write(nfecra,*) 'Plane averaging'
if((mod(ntcabs,ntchr).eq.0.and.ntchr.ne.-1.and.ntcabs.ge.4000).or.ntcabs.eq.ntmabs)then
   allocate(averages(nymax))

   write(nfecra,*) 'PLANE AVERAGING FOR',nymax, "cells"

   ! -- Loop on all the "moments" that contain averages
   do imom = 1, nbmomt
      write(nfecra,*) 'Averaging moment:',imom
      ipcmom = ipproc(icmome(imom))
      idtcm = -idtmom(imom)
      do ii = 1,nymax
         averages(ii) = 0.d0
      enddo
      ! -- Add the variables to the plane averaged arrays
      do iel  = 1, ncel
         ii = corrs(iel)
         averages(ii) = averages(ii)+propce(iel,ipcmom)                        
      enddo
      ! -- Sum the averages and the number of cells on all processors
      if(irangp.ge.0)then
         do ii = 1, nymax
            xabs = averages(ii)
            call parsom(xabs)
            averages(ii) = xabs
         enddo
      endif
      ! -- Obtain the averages by dividing by the number of cells on the streamwise & spanwise directions
      do ii = 1, nymax
         if(invarc(ii).le.0)then
            if(irangp.le.0) write(nfecra,*) 'RE-Warning ',ii, invarc(ii)           
         endif
         averages(ii) = averages(ii)/invarc(ii)
      enddo
      ! -- Add the averaged values to all cells to postprocess with ensight.
      do iel = 1, ncel
         ii = corrs(iel)
         propce(iel,ipcmom) = averages(ii)
      enddo
   enddo

   ! -- Calculate averaged gradients
   write(nfecra,*) 'Calculating the gradient of the averaged velocity'
   call flush(nfecra)
   do iel = 1,ncel
      imom = 1
      ipcmom = ipproc(icmome(imom))
      idtcm = -idtmom(imom)
      warr(iel) = propce(iel,ipcmom)/max(dtcmom(idtcm),epzero)
   enddo
   write(nfecra,*) 'Boundary Condition'
   ! Boundary condition pointers for gradients and advection
   call field_get_coefa_v(ivarfl(iu), coefav)
   call field_get_coefb_v(ivarfl(iu), coefbv)
   
   ! - Compute gradient
   
   write(nfecra,*) 'Gradient'
   call grdvec &
        !==========
        ( iu  , imrgra , inc    , iccocg ,                      &
        nswrgr(iu) , imligr(iu) ,                             &
        iwarni(iu) , nfecra ,                                 &
        epsrgr(iu) , climgr(iu) , extrag(iu) ,                &
        warr       , coefav , coefbv ,                        &
        grad  )
        
   !  Write a 1D file with all moments
   impout = impusr(1)
   write(nfecra,*) 'Write averages'
   if (irangp.le.0) then
      open(impout,file="averages.dat")
      write(impout,*) '# y(m),<U>,<V>,<W>,<uu>,<vv>,<ww>,<uv>,<uw>,<vw>,<p>,<du/dy> '         
   endif

   ymax = 2.d0
   icnt = 0

   npoint = 400
   iel1   = -999
   irang1 = -999
   do ii = 1, npoint
      xyz(1) = 0.d0
      xyz(2) = float(ii-1)/float(npoint-1)*ymax
      xyz(3) = 0.d0
      call findpt(ncelet, ncel, xyzcen,xyz(1), xyz(2), xyz(3), iel, irangv)

      if ((iel.ne.iel1).or.(irangv.ne.irang1)) then
         iel1   = iel
         irang1 = irangv
         
         if (irangp.eq.irangv) then
            do imom = 1, nbmomt
               ipcmom = ipproc(icmome(imom))
               idtcm = -idtmom(imom)
               datave(imom) = propce(iel,ipcmom)/max(dtcmom(idtcm),epzero)               
            enddo
            xdudy = grad(iel,2,1)
         else
            do imom = 1, nbmomt
               datave(imom) = 0.d0
            enddo
            xdudy = 0.d0
         endif
         if (irangp.ge.0) then
            iun = 1
            do imom = 1, nbmomt
               call parbcr(irangv,iun,datave(imom))
            enddo
            call parbcr(irangv,iun,xdudy)
         endif
         if (irangp.le.0)then
            icnt = icnt +1
            ! When writing, compute u'u' = <UU>-<U><U> and so on...
            ! assume the averages are as listed on the header.
            write(impout,1004) ydat(icnt),(datave(jj), jj = 1, 3)     &
                 ,datave(4) - datave(1)**2                              &
                 ,datave(5) - datave(2)**2                              &
                 ,datave(6) - datave(3)**2                              &
                 ,datave(7) - datave(1)*datave(2)                       &
                 ,datave(8) - datave(1)*datave(3)                       &
                 ,datave(9) - datave(2)*datave(3)                       &
                 ,(datave(jj), jj = 10, nbmomt)                         &
                 , xdudy
         endif
      endif
   enddo
   if (irangp.le.0) close(impout)
   deallocate(averages)   
endif
write(nfecra,*) 'Second Gradient'
! Boundary condition pointers for gradients and advection
call field_get_coefa_v(ivarfl(iu), coefav)
call field_get_coefb_v(ivarfl(iu), coefbv)

call grdvec &
     !==========
     ( iu  , imrgra , inc    , iccocg ,                      &
     nswrgr(iu) , imligr(iu) ,                             &
     iwarni(iu) , nfecra ,                                 &
     epsrgr(iu) , climgr(iu) , extrag(iu) ,                &
     rtpa(1,iu) , coefav , coefbv ,                        &
     grad  )

write(nfecra,*) 'Second Gradient finished'
dudy = 0.d0

! Add the variables to the plane averaged arrays
do iel  = 1, ncel
   ii = corrs(iel)
   dudy(ii) = dudy(ii)+grad(iel,2,1)        
enddo
! Sum the averages and the number of cells on all processors
if(irangp.ge.0)then
   do ii = 1, nymax
      xabs = dudy(ii)
      call parsom(xabs)
      dudy(ii) = xabs
   enddo
endif
! Obtain the averages by dividing by the number of cells on the streamwise & spanwise directions
do ii = 1, nymax
   dudy(ii) = dudy(ii)/invarc(ii)
enddo
write(nfecra,*) 'dudy finished'
write(nfecra,*) 'Time Series'
if (irangp.le.0) then
   impout = impusr(1)
   open(impout,file="TimeSeries.dat",access='append')
   write(impout,*) sqrt(abs(dudy(1))*viscl0)/viscl0,ntcabs*dtref
   close (impout)
endif

!--------
! Formats
!--------

1001 format('WARNING:  TOTAL NUMBER CELLS SHOULD BE',I5,/,       &
          'FOUND:',I5, 1E17.9)
1002 format(i3.3)
1003 format(2e17.9)
1004 format(20e17.9)

! Free memory
deallocate(grad)
deallocate(warr)
deallocate(dudy)

!----
! End
!----
return
end subroutine cs_user_extra_operations
